require "chipmunk"
require_relative "../graph_consts"
require_relative "../consts"
require_relative "../lib/bearlibmg"
require_relative "../lib/bearlibmap"

module TT
  FLOOR = 0
  WALL = 1
  NORTH = 2
  EAST = 3
  SOUTH = 4
  WEST = 5
  NE = 6
  SE = 7
  SW = 8
  NW = 9
  CORN_NE = 10
  CORN_SE = 11
  CORN_SW = 12
  CORN_NW = 13
  DOOR_N = 14
  DOOR_E = 15
  DOOR_S = 16
  DOOR_W = 17
end

module LogicalTile
  Wall = 0
  Room = 1
  Passage = 2
  VertDoor = 3
  HorizDoor = 4
  Stair = 5
end

def door?(x)
  [LogicalTile::VertDoor, LogicalTile::HorizDoor].include? x
end

class MapGenerator
  attr_accessor :size

  def initialize
    @double = true
    @tile_data = []
    @size = @double ? self.class.get_size / 2 : self.class.get_size
  end

  def tile(x, y)
    @tile_data[y * @size + x]
  end

  def double_size
    new_tile_data = Array.new(@size * @size * 4) do |i|
      x = i % (@size * 2)
      y = i / (@size * 2)
      @tile_data[y / 2 * @size + x / 2]
    end
    @tile_data = new_tile_data
    @size *= 2
  end

  def fix_walls
    (@size * @size).times do |i|
      x = i % @size
      y = i / @size
      next if x == @size - 1
      next if y == @size - 1
      curnw = Bearlib.passable? @tile_data[y * @size + x]
      curne = Bearlib.passable? @tile_data[y * @size + x + 1]
      cursw = Bearlib.passable? @tile_data[(y + 1) * @size + x]
      curse = Bearlib.passable? @tile_data[(y + 1) * @size + x + 1]
      if curnw && curse && !curne && !cursw
        if rand < 0.5
          ax = x + 1
          ay = y
        else
          ax = x
          ay = y + 1
        end
        @tile_data[ay * @size + ax] = @tile_data[y * @size + x]
      elsif !curnw && !curse && curne && cursw
        if rand < 0.5
          ax = x
          ay = y
        else
          ax = x + 1
          ay = y + 1
        end
        @tile_data[ay * @size + ax] = @tile_data[(y + 1) * @size + x]
      end
    end
  end

  def check_space
    @tile_data.count { |tile| Bearlib.passable?(tile) } >= 0.2 * @tile_data.size
  end

  def check_connected(point = nil)
    if point
      (@size * @size).times do |i|
        pass = Bearlib.passable?(@tile_data[i])
        @wave_data[i] = pass ? 0 : -1
      end
      @wave_data[point[1] * @size + point[0]] = 1
    else
      first = nil
      (@size * @size).times do |i|
        pass = Bearlib.passable?(@tile_data[i])
        @wave_data[i] = pass ? 0 : -1
        if !first && pass && !point
          first = i
          @wave_data[i] = 1
        end
      end
      return false unless first
    end
    #wave algorithm
    iter = 1
    loop do
      changed = false
      (@size * @size).times do |i|
        next unless @wave_data[i] == iter
        x = i % @size
        y = i / @size
        #paint neighbours
        [[0, 1], [0, -1], [1, 0], [-1, 0]].each do |dx, dy|
          ny = y + dy
          next if ny < 0 || ny >= @size
          nx = x + dx
          next if nx < 0 || nx >= @size
          ni = ny * @size + nx
          next if @wave_data[ni] != 0
          changed = true
          @wave_data[ni] = iter + 1
        end
      end
      break unless changed
      iter += 1
    end
    (@size * @size).times do |i|
      return false if @wave_data[i] == 0
    end
    true
  end

  private def get_tile(north, south, west, east, nw, ne, sw, se)
    return TT::NW if north && west
    return TT::NE if north && east
    return TT::SW if south && west
    return TT::SE if south && east
    return TT::NORTH if north
    return TT::SOUTH if south
    return TT::WEST if west
    return TT::EAST if east
    return TT::CORN_NW if nw
    return TT::CORN_NE if ne
    return TT::CORN_SW if sw
    return TT::CORN_SE if se
    return TT::FLOOR
  end
  private def get_door(north, south, west, east)
    return TT::DOOR_N if north
    return TT::DOOR_S if south
    return TT::DOOR_W if west
    return TT::DOOR_E if east
    return TT::FLOOR
  end

  def get_logical(cur, tile)
    return LogicalTile::Wall if tile == TT::WALL
    return LogicalTile::VertDoor if [TT::DOOR_N, TT::DOOR_S].include? tile
    return LogicalTile::HorizDoor if [TT::DOOR_W, TT::DOOR_E].include? tile
    return LogicalTile::Room if cur == Bearlib::TILE_BIGROOM
    return LogicalTile::Passage
  end

  def make_from_tile_data(data)
    (@size * @size).times do |i|
      x = i % @size
      y = i / @size

      cur = @tile_data[y * @size + x]
      if !Bearlib.passable?(cur) || x == 0 || x == @size - 1 || y == 0 || y == @size - 1
        tile = TT::WALL
      else
        north = !Bearlib.passable?(@tile_data[x + (y - 1) * @size])
        east = !Bearlib.passable?(@tile_data[(x + 1) + y * @size])
        west = !Bearlib.passable?(@tile_data[(x - 1) + y * @size])
        south = !Bearlib.passable?(@tile_data[x + (y + 1) * @size])
        nw = !Bearlib.passable?(@tile_data[(x - 1) + (y - 1) * @size])
        ne = !Bearlib.passable?(@tile_data[(x + 1) + (y - 1) * @size])
        sw = !Bearlib.passable?(@tile_data[(x - 1) + (y + 1) * @size])
        se = !Bearlib.passable?(@tile_data[(x + 1) + (y + 1) * @size])

        door = false
        if cur == Bearlib::TILE_DOOR
          if north || south
            door = @tile_data[(x + 1) + y * @size] == Bearlib::TILE_BIGROOM || @tile_data[(x - 1) + y * @size] == Bearlib::TILE_BIGROOM
          else
            door = @tile_data[x + (y + 1) * @size] == Bearlib::TILE_BIGROOM || @tile_data[x + (y - 1) * @size] == Bearlib::TILE_BIGROOM
          end
        end

        if door
          tile = get_door(north, south, west, east)
        else
          tile = get_tile(north, south, west, east, nw, ne, sw, se)
        end
      end
      data[i] = {
        pass: Bearlib.passable?(cur), #&& cur != Bearlib::TILE_DOOR,
        picture: BACK_TILES[tile.to_i],
        logic: get_logical(cur, tile),
      }
    end
  end
end

require_relative "dungeon"
require_relative "house"
require_relative "town"
require_relative "portaltown"

USE_MONSTERS = {
  HouseGenerator => 0,
  TownGenerator => 1,
  DungeonGenerator => 2,
  PortalTownGenerator => 3,
}

N_MONSTERS = {
  HouseGenerator => 20,
  TownGenerator => 300,
  DungeonGenerator => 100,
  PortalTownGenerator => 200,
}
